from __future__ import print_function

from rejson import Path
from functools import reduce
import json
import logging
from six import itervalues,iteritems
import sys

logger = logging.getLogger("reem")

SEPARATOR_CHARACTER = "&&&&"
ROOT_VALUE_SEQUENCE = "%%%%"

ROOT_PATH = Path.rootPath()


def path_to_key_sequence(path):
    """
    Turn a ReJSON subpath string into a sequence of key accesses
    :param path (str): a path string of form: ".", ".subkey1.subkey2"
    :return: list of key accesses below a top level key in redis
    :rtype: List[str]
    """
    if path == ROOT_PATH:
        return []
    return path.split(".")[1:]

def key_sequence_to_path(sequence):
    """
    Convert a sequence of key accesses into a path string representing a path below the top level key in redis
    :param sequence (List[str]): list of strings representing key accesses
    :return: a subpath string
    """
    return ROOT_PATH + ".".join(sequence)

def key_sequence_to_path_ext(sequence):
    """
    Convert a sequence of key accesses or array accesses into a path string representing a path below the top level key in redis
    :param sequence (List[str or int]): list of strings representing key accesses or indices representing array accesses
    :return: a subpath string
    """
    if len(sequence) == 0:
        return ROOT_PATH
    #if isinstance(sequence[0],int):
    #    raise ValueError("Top-level key cannot be an integer")
    seps_keys = []
    for k in sequence:
        if isinstance(k,int):
            seps_keys.append('[%d]'%(k,))
        else:
            seps_keys.append('.')
            seps_keys.append(k)
    return ''.join(seps_keys)


def copy_dictionary_without_paths(dictionary, key_sequence):
    """
    Return a copy of dictionary with the paths formed by key_sequences removed
    :param dictionary (Dict): The dictionary to copy
    :param key_sequence (List[List[str]]): 2D List of Strings where each internal list is a sequence of key accesses from dictionary root
    :return: dictionary with paths represented by lists inside key_sequences removed
    :rtype: dict
    """
    ret = {}
    possibles = [ks for ks in key_sequence if len(ks) == 1]
    possibles = set(reduce(lambda x, y: x + y, possibles, []))
    for k, v in iteritems(dictionary):
        if k in possibles:
            continue
        if type(v) == dict:
            ret[k] = copy_dictionary_without_paths(v, [ks[1:] for ks in key_sequence if len(ks) > 1])
        else:
            ret[k] = v
    return ret


def extract_object(dictionary, key_sequence):
    """
    Extract the object inside the dictionary at a specific path
    :param dictionary (Dict): dictionary to extract from
    :param key_sequence (List[str]): list of strings represetning key accesses
    :return: the value inside dictionary specified by key_sequence
    """
    ret = dictionary
    for k in key_sequence:
        ret = ret[k]
    return ret


def filter_paths_by_prefix(paths, prefix):
    """
    Given a list of paths and a prefix, return the paths that start with prefix.
    :param paths (Iterable[str]): list of paths
    :param prefix (str): prefix to filter on
    :return: list of paths that begin with the specified prefix and the subsequent paths after the prefix
    :rtype: List[str], List[str]
    """
    if prefix == ROOT_PATH:
        paths = list(paths)
        return paths, paths
    full_paths, suffixes = [], []
    for p in paths:
        if p.startswith(prefix):
            suffixes.append(p[len(prefix):])
            full_paths.append(p)
    return full_paths, suffixes


def insert_into_dictionary(dictionary, key_sequence, value):
    """
    Insert value into dictionary at path specified by key_sequence. If the sequence of keys does not exist, it will be made
    :param dictionary (Dict): dictionary to insert into
    :param key_sequence (List[str]): specify the path inside the dictionary
    :param value: value to insert
    :return: None - the method will modify dictionary
    :rtype: None
    """
    parent = dictionary
    for key in key_sequence[:-1]:
        parent = parent.setdefault(key,{})
    parent[key_sequence[-1]] = value


def get_special_paths(set_path, set_value, sp_to_label, label_to_ship):
    """
    Get info on how to update sp_to_label according to the user uploading `set_value` in location `set_path`
    :param set_path (str): ".key0.key1" subpath
    :param set_value: anything to be inserted into data. Needs to uploadable to redis
    :param sp_to_label (Dict): dictionary mapping current special paths to their ships
    :param label_to_ship (Dict): dictionary mapping ship labels to ship objects
    :return: A set of tuples: (path, corresponding ship's label)
    :rtype: set
    """
    additions = set()
    if type(set_value) is not dict:
        # If this path is already labelled as special, then check that the ship matched with it does fit
        if set_path in sp_to_label:
            if label_to_ship[sp_to_label[set_path]].check_fit(set_value):
                return additions

        # If this is something a ship covers, add it as a special path
        for ship in itervalues(label_to_ship):
            if ship.check_fit(set_value):
                additions.add( (set_path, ship.get_label()) )

    # If this is a dict, recursively build the set of additional paths
    else:
        for k, v in iteritems(set_value):
            assert check_valid_key_name(k), "Invalid Key: {}".format(k)
            if set_path != ROOT_PATH:
                child_path = "{}.{}".format(set_path, k)
            else:
                child_path = ".{}".format(k)
            child_add = get_special_paths(child_path, v, sp_to_label, label_to_ship)
            additions = additions.union(child_add)

    return additions


def check_valid_key_name(name):
    """
    Ensure the key name provided is legal
    :param name: a potential key name
    :return: boolean indicating legality of name
    :rtype: bool
    """
    if not isinstance(name,str):
        return False
    bad_chars = ["*", ".", "&&&&"]
    if any(k in name for k in bad_chars):
        return False
    return True

def check_valid_key_name_ext(name):
    """
    Ensure the key / index name provided is legal
    :param name: a potential key name or index
    :return: boolean indicating legality of name
    :rtype: bool
    """
    if isinstance(name,int):
        return True
    if not isinstance(name,str):
        return False
    bad_chars = ["*", ".", "&&&&"]
    if any(k in name for k in bad_chars):
        return False
    return True

if sys.version_info[0] < 3:
    def json_recode_str(input):
        if isinstance(input, dict):
            return {json_recode_str(key): json_recode_str(value)
                    for key, value in input.iteritems()}
        elif isinstance(input, list):
            return [json_recode_str(element) for element in input]
        elif isinstance(input, unicode):
            return input.encode('utf-8')
        else:
            return input
else:
    def json_recode_str(input):
        return input


__version__ = '0.1.0'
__all__ = ['RedisInterface','KeyValueStore','PublishSpace','CallbackSubscriber','SilentSubscriber']
from .connection import RedisInterface
from .datatypes import KeyValueStore,PublishSpace,SilentSubscriber,CallbackSubscriber


from __future__ import print_function
from six import iterkeys
import builtins

from .utilities import *
from .supports import KeyAccessor, WriteOnlyKeyAccessor, ChannelListener, ActiveSubscriberKeyAccessor
from rejson import Path
from threading import Lock
import logging
import queue
from . import RedisInterface

logger = logging.getLogger("reem")

_ROOT_VALUE_READ_NAME = "{}ROOT{}".format(ROOT_VALUE_SEQUENCE, ROOT_VALUE_SEQUENCE)

class KeyValueStore(object):
    """Dictionary Like object used for get/set paradigm

    The ``KeyValueStore`` object is one that users will frequently use. It
    keeps ``Reader`` and ``Writer`` objects for each key that the user has read
    or written to. It does not actually handle the getting and setting of data
    but produces ``KeyAccessor`` objects that assist with path
    construction and call the reader and writer's write and read methods.

    Attributes:
        interface (str or RedisInterface): Defines the connection to Redis this
        reader will use. If a str, then a RedisInterface will be created and
        connected to automatically.
    """
    def __init__(self, interface='localhost'):
        if isinstance(interface,str):
            host = interface
            self.interface = RedisInterface(host)
            self.interface.initialize()
        elif isinstance(interface,KeyValueStore):
            self.interface = interface.interface
            assert isinstance(self.interface,RedisInterface)
        else:
            assert isinstance(interface,RedisInterface)
            self.interface = interface
        self.entries = {}
        self.track_schema = True
    
    def get(self,path):
        """Reads a path in the form 'key.element1.element2' from the redis
        server.  If the path doesn't exist, an error is raised.

        Marginally faster than nested accessors.
        """
        res = path.split('.',1)
        if len(res) == 1:
            res = [res[0],'']
        if res[0] not in self.entries:
            raise ValueError("Invalid top-level key",res[0])
        writer,reader = self.entries[res[0]]
        return reader.read_from_redis('.'+res[1])

    def set(self,path,value):
        """Sets a value in a path in the form 'key.element1.element2' from the
        redis server.  If any elements along the path, except for the last key,
        do not exist, an error is raised.

        Marginally faster than nested accessors.
        """
        res = path.split('.',1)
        if len(res) == 1:
            res = [res[0],'']
        if res[0] not in self.entries:
            raise ValueError("Invalid top-level key",res[0])
        writer,reader = self.entries[res[0]]
        return writer.send_to_redis('.'+res[1],value)
        
    def __setitem__(self, key, value):
        """ Only used for setting key on first level of KVS. i.e. KVS["top_key"] = value. Otherwise see __getitem__

        Args:
            key (str): "dictionary" key. It becomes a top-level Redis key
            value: value to store

        Returns: None

        """

        assert check_valid_key_name(key), "Invalid Key: {}".format(key)
        if not isinstance(value,(list,tuple,dict)):
            value = {_ROOT_VALUE_READ_NAME: value}
        with self.interface.INTERFACE_LOCK:
            self.__ensure_key_existence(key)
            writer, reader = self.entries[key]
        writer.send_to_redis(Path.rootPath(), value)

    def __getitem__(self, item):
        """Used to retrieve KeyAccessor object for handling path construction when setting/getting Redis

        Args:
            item (str): "dictionary" key. It must be a top-level Redis key

        Returns: KeyAccessor

        """

        assert check_valid_key_name(item), "Invalid Key: {}".format(item)
        #logger.debug("KVS GET {}".format(item))
        with self.interface.INTERFACE_LOCK:
            self.__ensure_key_existence(item)
            writer, reader = self.entries[item]
        return KeyAccessor(self, writer=writer, reader=reader)

    def __delitem__(self,item):
        """ Only used for deleting key on first level of KVS. 

        Args:
            key (str): "dictionary" key. It must be a top-level Redis key

        Returns: None

        """
        with self.interface.INTERFACE_LOCK:
            if item not in self.entries:
                raise KeyError(item)
            writer, reader = self.entries[item]
        writer.delete_from_redis(Path.rootPath())

    def __ensure_key_existence(self, key):
        """ Ensure that the requested key has a reader and writer associated with it.

        Returns: None

        """

        assert check_valid_key_name(key), "Invalid Key: {}".format(key)    
        if key not in self.entries:
            self.entries[key] = (Writer(key, self.interface), Reader(key, self.interface))
            self.entries[key][0].do_metadata_update = self.track_schema

    def track_schema_changes(self, set_value, keys=None):
        """ Performance optimization for skipping schema update checks

        Stop checking for schema updates when setting data. Use ONLY if your data's schema is static
        and you are really trying to eek out every bit of optimization.

        Args:
            set_value (bool): True/False indicating if the keys' schema should be tracked
            keys (List[str]): List of keys to track. If None, all present and future keys are tracked
            according to ``set_value``

        Returns: None

        """
        with self.interface.INTERFACE_LOCK:
            if keys is None:
                keys = iterkeys(self.entries)
                self.track_schema = set_value
            for k in keys:
                self.entries[k][0].do_metadata_update = set_value



class PublishSpace(KeyValueStore):
    """ Convenience class for publishing

    This class keeps track of ``PublishWriter`` objects for each key the user has published on.

    """
    def __getitem__(self, item):
        """ Same function as ``KeyValueStore`` method, but returns a non-readable ``WriteOnlyKeyAccessor`` object

        """
        assert isinstance(item,str)
        with self.interface.INTERFACE_LOCK:
            self.__ensure_key_existence(item)
            publisher, _ = self.entries[item]
        return WriteOnlyKeyAccessor(self, writer=publisher, reader=_)

    def __ensure_key_existence(self, key):
        """ Identical to ``KeyValueStore`` method of same name but does not instantiate a ``Reader`` object
        """

        assert check_valid_key_name(key), "Invalid Key: {}".format(key)
        if key not in self.entries:
            self.entries[key] = (PublishWriter(key, self.interface), None)
            self.entries[key][0].do_metadata_update = self.track_schema


class RawSubscriber:
    def __init__(self, channel_name, interface, callback_function, kwargs):
        self.listening_channel = '__pubspace@0__:{}'.format(channel_name)
        self.listener = ChannelListener(interface, self.listening_channel, callback_function, kwargs)
        self.listener.setDaemon(True)

    def listen(self):
        self.listener.start()


class SilentSubscriber:
    """ Silent Subscriber Implementation"""

    def __init__(self, channel, interface):
        if isinstance(interface,str):
            host = interface
            interface = RedisInterface(host)
            interface.initialize()
        self.passive_subscriber = RawSubscriber(channel + "*", interface, self.update_local_copy, {})
        self.reader = Reader(channel, self.passive_subscriber.interface)
        self.local_copy = {}
        self.prefix = "__pubspace@0__:" + self.reader.top_key_name

    def update_local_copy(self, channel, message):
        """ Update the local copy of the data stored under this channel name in redis.

        Args:
            channel: the name of the channel that was published.
            message: message published on that channel

        Returns: None

        """

        #logger.info("SILENT_SUBSCRIBER @{} : channel={} message={}".format(self.prefix, channel, message))
        try:
            message = message.decode("utf-8")
        except Exception as e:
            return
        if message != "Publish":
            return

        if channel == self.prefix:
            self.local_copy = self.reader.read_from_redis(Path.rootPath())
            return

        path = channel[len(self.prefix):]
        redis_value = self.reader.read_from_redis(path)
        #logger.debug("SILENT_SUBSCRIBER @{} : Read from Redis: {}".format(self.prefix, redis_value))
        insert_into_dictionary(self.local_copy, path_to_key_sequence(path), redis_value)
        #logger.debug("SILENT_SUBSCRIBER @{} : Local Copy: {}".format(self.prefix, self.local_copy))

    def listen(self):
        """ Makes this subscriber start listening

        Returns: None

        """
        self.passive_subscriber.listen()

    def value(self):
        """ Get data stored at root

        Where as the reader can do ``server["foo"].read()`` with if server is a ``KeyValueStore``,
        accessing the root value of a subscriber is not as easy. This method retrieves all data stored underneath
        a top level key.

        Returns: all data stored underneath a top level Redis key.

        """
        if _ROOT_VALUE_READ_NAME in self.local_copy:
            return self.local_copy[_ROOT_VALUE_READ_NAME]
        # Copy dictionary - paths to omit is blank, so we copy everything
        return copy_dictionary_without_paths(self.local_copy, [])

    def __getitem__(self, item):
        """ Implement dictionary API for local copy

        We do not give the user direct access to the dictionary representing the lcoal
        Args:
            item:

        Returns: ActiveSubscriberKeyAccessor

        """
        assert isinstance(item,str), "Key name must be string"
        return ActiveSubscriberKeyAccessor(self, None, self.reader, [item])


class CallbackSubscriber(SilentSubscriber):
    """ Callback Subscriber Implementation"""

    def __init__(self, channel, interface, callback_function, kwargs):
        if isinstance(interface,str):
            host = interface
            interface = RedisInterface(host)
            interface.initialize()
        super(CallbackSubscriber,self).__init__(channel, interface)
        self.passive_subscriber = RawSubscriber(channel + "*", interface, self.call_user_function, {})
        self.callback_function = callback_function
        self.kwargs = kwargs

    def call_user_function(self, channel, message):
        """
        Wrapper callback function (wrapping user function) for this class to work with a RawSubscriber object
        Fits required interface for a ChannelSubscriber callback function
        :param channel: channel published to
        :param message: message that was published
        :return: None
        :rtype: None
        """
        self.update_local_copy(channel, message)
        channel_name = channel.split("__pubspace@0__:")[1]
        self.callback_function(data=self.value(), updated_path=channel_name, **self.kwargs)






class Writer(object):
    """Responsible for setting data inside Redis

    The Writer class is an internal class that is used for all data sent to Redis (not including pub/sub messages).
    Each key that will have nested data below requires a new instantiation of Writer

    Attributes:
        top_key_name (str): The name of the Redis key under which JSON data will be stored. To Redis,
            this will become a ReJSON key name. It is also used to generate the Redis key name that ``ship``'s use to
            store non JSON data.
        interface (RedisInterface): Defines the connection to Redis this writer will use
    """
    def __init__(self, top_key_name, interface):
        self.interface = interface
        self.top_key_name = top_key_name
        self.metadata_key_name = self.top_key_name + SEPARATOR_CHARACTER + "metadata"
        self.metadata = None
        self.__initialize_metadata()
        self.sp_to_label = self.metadata["special_paths"]
        self.pipeline = self.interface.client.pipeline()
        self.pipeline_no_decode = self.interface.client_no_decode.pipeline()
        self.do_metadata_update = True

    def __initialize_metadata(self):
        """ Pull metadata for this key from Redis or set a default

        Returns:
            None
        """
        try:
            pulled = self.interface.client.jsonget(self.metadata_key_name, ".")
            if pulled is not None:
                self.metadata = pulled
                return
        except TypeError:
            pass
        self.metadata = {"special_paths": {}, "required_labels": self.interface.shipper_labels}

    def send_to_redis(self, set_path, set_value):
        """ Execute equivalent of ``JSON.SET self.top_key_name <set_path> <set_value>``

        From the user's perspective, it executes ``JSON.SET self.top_key_name <set_path> <set_value>``
        except that ``set_value`` can be json-incompatible.  In such a case, it determines what
        non-serializable types are inside set_value, stores the serializable data as a JSON, and
        stores the non-serializable data using ``self.interface``'s ships.

        Args:
            set_path (str): path underneath JSON key to set
            set_value: value to set

        Returns:
            None

        """
        with self.interface.INTERFACE_LOCK:
            #logger.info("SET {} {} = {}".format(self.top_key_name, set_path, type(set_value)))
            self.__process_metadata(set_path, set_value)
            #logger.debug("SET {} {} Metadata: {}".format(self.top_key_name, set_path, self.metadata))
            self.__publish_non_serializables(set_path, set_value)
            #logger.debug("SET {} {} Non-Serializables Pipelined".format(self.top_key_name, set_path))
            self.__publish_serializables(set_path, set_value)
            #logger.debug("SET {} {} Serializables Pipelined".format(self.top_key_name, set_path))
            self.pipeline.execute()
            #logger.debug("SET {} {} Pipeline Executed".format(self.top_key_name, set_path))

    def delete_from_redis(self,del_path):
        """Execute equivalent of ``JSON.DEL self.top_key_name <set_path>``


        Args:
            del_path (str): path underneath JSON key to delete

        Returns:
            None
        """
        with self.interface.INTERFACE_LOCK:
            dels = 0
            if del_path in self.sp_to_label:
                self.pipeline_no_decode = self.interface.client_no_decode.pipeline()
                shipper = self.interface.label_to_shipper[self.sp_to_label[del_path]]
                special_name = str(self.top_key_name) + str(del_path)
                shipper.delete(special_name, self.pipeline_no_decode)
                self.pipeline_no_decode.execute()
                del self.sp_to_label[del_path]
                dels += 1
            else:
                self.pipeline.jsondel(self.top_key_name, del_path)

                #maybe some of the sub-keys of del_path could be special
                special_paths, suffixes = filter_paths_by_prefix(iterkeys(self.sp_to_label), del_path)
                for p in special_paths:
                    special_path_redis_key_name = self.top_key_name + p
                    ship = self.interface.label_to_shipper[self.sp_to_label[p]]
                    ship.delete(special_path_redis_key_name, self.pipeline_no_decode)
                    del self.sp_to_label[p]
                    dels += 1
                self.pipeline_no_decode.execute()

            if dels > 0:
                #must broadcast metadata change
                self.pipeline.jsonset(self.metadata_key_name, ".", self.metadata)
                channel, message = "__keyspace@0__:"+ self.metadata_key_name, "set"
                self.pipeline.publish(channel, message)  # Homemade key-space notification for metadata updates

            self.pipeline.execute()


    def __process_metadata(self, set_path, set_value):
        """ Handle metadata updates

        Given the path and value the user would like to set, check if there are non-serializable data types and update
        metadata locally and in Redis. Happens without pipeline

        Args:
            set_path (str): path underneath JSON key to set
            set_value: value to set

        Returns:
            None

        """
        if not self.do_metadata_update:
            return

        overridden_paths = set()
        for existing_path in iterkeys(self.sp_to_label):
            if existing_path.startswith(set_path):
                overridden_paths.add(existing_path)
        [self.sp_to_label.pop(op) for op in overridden_paths]

        special_paths = get_special_paths(set_path, set_value, self.sp_to_label, self.interface.label_to_shipper)
        dels, adds = overridden_paths - special_paths, special_paths - overridden_paths
        for set_path, label in special_paths:
            self.sp_to_label[set_path] = label

        if len(adds) > 0 or len(dels) > 0:
            self.pipeline.jsonset(self.metadata_key_name, ".", self.metadata)
            channel, message = "__keyspace@0__:"+ self.metadata_key_name, "set"
            self.pipeline.publish(channel, message)  # Homemade key-space notification for metadata updates

    def __publish_non_serializables(self, set_path, set_value):
        """Publish JSON incompatible data to Redis

        Given a set, publish the non-serializable components to redis, given that metadata has been updated already

        Args:
            set_path (str): path underneath JSON key to set
            set_value: value to set

        Returns:
            None
        """

        overridden_paths, suffixes = filter_paths_by_prefix(iterkeys(self.sp_to_label), set_path)
        for full_path, suffix in zip(overridden_paths, suffixes):
            #logger.debug("Suffix = {}".format(suffix))
            update_value = extract_object(set_value, path_to_key_sequence(suffix))
            special_path_redis_key_name = self.top_key_name+full_path
            #logger.debug("SET {} {} Non-serializable update {} = {}".format(
            #    self.top_key_name, set_path, special_path_redis_key_name, type(update_value))
            #)
            self.interface.label_to_shipper[self.sp_to_label[full_path]].write(
                key=special_path_redis_key_name,
                value=update_value,
                client=self.pipeline
            )

    def __publish_serializables(self, set_path, set_value):
        """ Publish the serializable portion of ``set_value

        Take out the non-serializable part of set_value and publish it at set_path

        Args:
            set_path (str): path underneath JSON key to set
            set_value: value to set

        Returns: None

        """
        if type(set_value) is dict:
            intrusive_paths, suffixes = filter_paths_by_prefix(iterkeys(self.sp_to_label), set_path)
            excised_copy = copy_dictionary_without_paths(set_value, [path_to_key_sequence(s) for s in suffixes])
            self.pipeline.jsonset(self.top_key_name, set_path, excised_copy)
            #logger.debug("SET {} {} Serializable update {} = {}".format(self.top_key_name, set_path, set_path, excised_copy))
        elif set_path not in self.sp_to_label:
            self.pipeline.jsonset(self.top_key_name, set_path, set_value)
            #logger.debug("SET {} {} Serializable update {} = {}".format(self.top_key_name, set_path, set_path, set_value))


class Reader(object):
    """Responsible for getting data from Redis

    The Reader class is an internal class that is used for all read from Redis (not including pub/sub messages).
    Each key that will have nested data below requires a new instantiation of Reader

    Attributes:
        top_key_name (str): The name of the Redis key under which JSON data is stored
        interface (RedisInterface): Defines the connection to Redis this reader will use
    """
    def __init__(self, top_key_name, interface):
        self.interface = interface
        self.top_key_name = top_key_name
        self.metadata = {"special_paths": {}, "required_labels": self.interface.shipper_labels}
        self.sp_to_label = self.metadata["special_paths"]
        self.pipeline = self.interface.client.pipeline()
        self.pipeline_no_decode = self.interface.client_no_decode.pipeline()
        self.metadata_key_name = "{}{}metadata".format(self.top_key_name, SEPARATOR_CHARACTER)
        self.interface.metadata_listener.add_listener(self.metadata_key_name, self)
        self.pull_metadata = True
        # Will need to update metadata on first read regardless so the simple initialization we have here is sufficient

    def read_from_redis(self, read_path):
        """ Read specified path from Redis

        This is the only public method of the Reader class. It will retrieve the data stored at a specified path
        from Redis. At a high level, it reads data stored with ReJSON and inserts non-JSON compatible data
        at appropriate paths using the metadata associated with this key.

        Args:
            read_path (str): path the user wants to read

        Returns: data stored at value in Redis

        """

        with self.interface.INTERFACE_LOCK:
            #logger.info("GET {} {} pull_metadata = {}".format(self.top_key_name, read_path, self.pull_metadata))
            self.update_metadata()
            #logger.debug("GET {} {} Using Metadata: {}".format(self.top_key_name, read_path, self.metadata))
            if read_path in self.sp_to_label:
                ret_val = self.pull_special_path(read_path)
            else:
                self.queue_reads(read_path)
                #logger.debug("GET {} {} Reads Queued".format(self.top_key_name, read_path))
                ret_val = item = json_recode_str(self.build_dictionary(read_path))
                #logger.debug("GET {} {} Dictionary Built".format(self.top_key_name, read_path))
        return ret_val

    def update_metadata(self):
        """ Update the local copy of metadata if a relevant path has been updated.

        The metadata listener is a redis client subscribed to key-space notifications. If a relevant path is updated,
        this Reader's ``pull_metadata`` flag will be turned on. If ``pull_metadata`` is ``True``, then the reader
        will fetch metadata from the Redis server.

        Returns: None

        """
        self.interface.metadata_listener.flush()
        if self.pull_metadata:
            try:
                pulled = self.interface.client.jsonget(self.metadata_key_name, ".")
                if pulled is not None:
                    self.metadata = pulled
            except TypeError:  # No Metadata online
                return
            self.sp_to_label = self.metadata["special_paths"]
        self.pull_metadata = False

    def queue_reads(self, read_path):
        """ Queue reads in a pipeline

        Queue all redis queries necessary to read data at path into the appropriate redis pipeline.
        First, queue decoded pipeline with the ReJSON query
        Next, queue all the special path reads with the non-decoded pipeline and ships

        Args:
            read_path: path user wants to read

        Returns: None
        """
        self.pipeline.jsonget(self.top_key_name, read_path)
        special_paths, suffixes = filter_paths_by_prefix(iterkeys(self.sp_to_label), read_path)
        for p in special_paths:
            special_path_redis_key_name = self.top_key_name + p
            #logger.debug("type(sp to label) = {}, type(p) = {}".format(type(self.sp_to_label), type(p)))
            ship = self.interface.label_to_shipper[self.sp_to_label[p]]
            ship.read(special_path_redis_key_name, self.pipeline_no_decode)

    def build_dictionary(self, read_path):
        """ Execute pipelines and consolidate data into a dictionary

        Args:
            read_path: path user wants to read

        Returns: The data stored at ``read_path`` in Redis
        """

        try:
            return_val = self.pipeline.execute()[0]
        except TypeError as e:
            raise KeyError("Invalid read from path "+self.top_key_name+read_path)
        #logger.debug("GET {} {} Serializable Pipeline Executed".format(self.top_key_name, read_path))
        responses = self.pipeline_no_decode.execute()
        special_paths, suffixes = filter_paths_by_prefix(iterkeys(self.sp_to_label), read_path)
        assert len(special_paths) == len(suffixes)
        #logger.debug("special_path = {}, suffixes = {}".format(special_paths, suffixes))
        for i, (sp, suffix) in enumerate(zip(special_paths, suffixes)):
            value = self.interface.label_to_shipper[self.sp_to_label[sp]].interpret_read(responses[i: i + 1])
            insert_into_dictionary(return_val, path_to_key_sequence(suffix), value)
            #logger.debug("GET {} {} Nonserializable Pipeline Inserted {} = {}"
            #             .format(self.top_key_name, read_path, sp, type(value))
            #             )
        #logger.debug("GET {} {} Dictionary Built".format(self.top_key_name, read_path))
        return return_val

    def pull_special_path(self, read_path):
        """ Directly pull a non-JSON path

        If the user specified path is not in JSON, this will retrieve the data directly without going through ReJSON.

        Args:
            read_path: path user wants to read

        Returns:

        """

        shipper = self.interface.label_to_shipper[self.sp_to_label[read_path]]
        special_name = str(self.top_key_name) + str(read_path)
        shipper.read(special_name, self.pipeline_no_decode)
        responses = self.pipeline_no_decode.execute()
        return shipper.interpret_read(responses)


class PublishWriter(Writer):
    """ Defines PublishWriter behavior

    The PublishWriter is identical to the writer but publishes a message when it writes a value.

    """

    def __init__(self, top_key_name, interface):
        super(PublishWriter,self).__init__(top_key_name, interface)
        self.message = "Publish"

    def send_to_redis(self, set_path, set_value):
        """ PublishWriter equivalent of Writer ``send_to_redis``

        This is an equivalent function to ``Writer``'s ``send_to_redis`` method but also publishes a message
        indicating what channel has been updated.

        Args:
            set_path (str): path underneath JSON key to set
            set_value: value to set

        Returns: None

        """

        #logger.info("PUBLISH {} {} = {}".format(self.top_key_name, set_path, type(set_value)))
        #logger.debug("PUBLISH {} {} Metadata Update?: {}".format(self.top_key_name, set_path, self.do_metadata_update))
        with self.interface.INTERFACE_LOCK:
            self.__process_metadata(set_path, set_value)
            #logger.debug("PUBLISH {} {} Metadata: {}".format(self.top_key_name, set_path, self.metadata))
            self.__publish_non_serializables(set_path, set_value)
            self.__publish_serializables(set_path, set_value)

            # Addition to Writer class
            if set_path == Path.rootPath():
                set_path = ""
            channel_name = "__pubspace@0__:" + self.top_key_name + set_path
            self.pipeline.publish(channel_name, self.message)

            # Resume Writer Class
            self.pipeline.execute()
            #logger.debug("PUBLISH {} {} pipeline executed, published {} to {}".format(
            #    self.top_key_name, set_path, self.message, channel_name)
            #)

    def delete_from_redis(self,del_path):
        raise RuntimeError("Cannot delete published values")


from __future__ import print_function,unicode_literals

from abc import ABCMeta, abstractmethod
import numpy as np
import io


class SpecialDatatypeShip(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        super(SpecialDatatypeShip, self).__init__()

    @abstractmethod
    def check_fit(self, value):
        """ Determine if this ship will handle ``value``

        This method returns true if ``value`` is data that this ship is supposed to handle. If this ship handled all
        numpy arrays, it would check if ``value``'s type is a numpy array.

        Args:
            value: object to check

        Returns: True if ship will handle ``value``

        """
        pass

    @abstractmethod
    def write(self, key, value, client):
        """ Write ``value`` to Redis at the specified ``key`` using ``client``

        Given a Redis client, execute any number of needed commands to store the ``value`` in Redis. You
        are required to use the key given for REEM to find it. If you must store multiple pieces of information,
        use a `Redis Hash <https://redis.io/topics/data-types>`_ which acts like a one level dictionary.

        Args:
            key (str): The Redis key name this ship must store data under
            value: The value to write into Redis
            client: A `ReJSON Redis Client <https://github.com/RedisJSON/rejson-py>`_ pipeline

        Returns: None

        """
        pass

    @abstractmethod
    def read(self, key, client):
        """ Retrieve necessary information from Redis

        Given a Redis client, execute ONE command to retrieve all the information you need to rebuild the data
        that was stored in ``write`` from Redis. This method should execute the command that allows you to retrieve
        all data stored under key

        Args:
            key (str): a keyname that contains data stored by ``write``
            client: A `ReJSON Redis Client <https://github.com/RedisJSON/rejson-py>`_ pipeline

        Returns: None

        """
        pass

    @abstractmethod
    def delete(self, key, client):
        """Delete information from redis at the specified ``key`` using ``client``.

        Given a Redis client, execute any number of commands to retrieve all the information you need to 
        delete all of the data stored under ``key``

        Args:
            key (str): a keyname that contains data stored by ``write``
            client: A `ReJSON Redis Client <https://github.com/RedisJSON/rejson-py>`_ pipeline

        Returns: None

        """
        pass

    @abstractmethod
    def interpret_read(self, responses):
        """ Translate Redis data into a local object

        Redis will reply to you with something according to what read command you executed in ``read``. This method
        takes whatever Redis replied with and turns it into an object identical to what was initially passed to
        ``write`` as value.

        Args:
            responses: Redis's reply data based on ``read`` method

        Returns: An object identical to what was initially written to Redis.

        """
        pass

    @abstractmethod
    def get_label(self):
        """ Return a unique string identifier

        This method should return a string that uniquely identifies this ship. REEM will use it to determine what ship
        to use to decode data that is already stored in Redis.

        Returns:
             str: the string identifier

        """
        pass


class NumpyShip(SpecialDatatypeShip):
    def check_fit(self, value):
        return type(value) in [np.array, np.ndarray]

    def get_label(self):
        return "default_numpy_handler"

    def write(self, key, value, client):
        client.hset(key, "arr", memoryview(value.data).tobytes())
        client.hset(key, "dtype", str(value.dtype))
        client.hset(key, "shape", str(value.shape))
        client.hset(key, "strides", str(value.strides))

    def delete(self, key, client):
        client.delete(key)

    def read(self, key,  client):
        client.hgetall(key)

    def interpret_read(self, responses):
        hash = responses[0]
        dtype = eval("np.{}".format(hash[b'dtype'].decode('utf-8')))
        shape = hash[b'shape'].decode("utf-8")[1:-1]
        shape = tuple([int(s) for s in shape.split(",") if len(s) > 0])
        arr = np.frombuffer(hash[b'arr'], dtype)
        arr = np.reshape(arr, shape)
        return arr



from __future__ import print_function

import rejson
from .supports import  MetadataListener
from .ships import *
from threading import Lock


class RedisInterface:
    """

    """
    def __init__(self, host='localhost', ships=[NumpyShip()]):
        self.hostname = host
        self.ships = ships
        self.client = rejson.Client(host=host, decode_responses=True)
        self.client_no_decode = rejson.Client(host=host)
        self.metadata_listener = MetadataListener(self)
        self.INTERFACE_LOCK = Lock()

        self.label_to_shipper = {}
        for sh in self.ships:
            self.label_to_shipper[sh.get_label()] = sh

        self.shipper_labels = [sh.get_label() for sh in ships]

    def initialize(self):
        """
        Call before doing anything with this connection
        :return:
        """
        pass


from __future__ import print_function

from threading import Thread, Lock
import rejson
from .utilities import *

_ROOT_VALUE_READ_NAME = "{}ROOT{}".format(ROOT_VALUE_SEQUENCE, ROOT_VALUE_SEQUENCE)
_TYPEMAP = {
    'object':dict,
    'array':list,
    'integer':int,
    'number':float,
    'boolean':bool
}

class MetadataListener:
    def __init__(self, interface):
        self.client = rejson.Client(host=interface.hostname)
        self.pubsub = self.client.pubsub()
        self.pubsub.psubscribe(['__keyspace@0__:*'])
        self.listeners = {}

    def add_listener(self, key_name, reader):
        listen_name = "__keyspace@0__:{}".format(key_name)
        self.listeners.setdefault(listen_name,[]).append(reader)

    def flush(self):
        while True:
            item = self.pubsub.get_message()
            if item is None:
                break
            channel = item['channel'].decode("utf_8")
            try:
                for listener in self.listeners[channel]:
                    listener.pull_metadata = True
            except KeyError:
                pass


class KeyAccessor:
    """Main class for accessing sub-keys of a KeyValueStore."""
    def __init__(self, parent, writer, reader, initial_path=[]):
        self.parent = parent
        self.writer = writer
        self.reader = reader
        self.path = initial_path
        self.path_str = None  #caches path string

    def __str__(self):
        if self.path_str is None:
            self.path_str = key_sequence_to_path_ext(self.path)
        return "reem.KeyAccessor({} {})".format(self.writer.top_key_name,self.path_str)

    def __getitem__(self, key):
        assert check_valid_key_name_ext(key),"{} is not a valid key under path {}".format(key,key_sequence_to_path(self.path) if self.path_str is None else self.path_str)
        return self.__class__(self,self.writer,self.reader,self.path+[key])

    def __setitem__(self, key, value):
        if isinstance(value,KeyAccessor):
            #sometimes this happens on += / *= on subkeys
            if value.parent is self:
                return
            else:
                raise ValueError("Cannot set a KeyAccessor to another KeyAccessor... {}[{}] = {}".format(self.path,key,value.path))
        assert check_valid_key_name_ext(key),"{} is not a valid key under path {}".format(key,key_sequence_to_path(self.path) if self.path_str is None else self.path_str)
        if self.path_str is None:
            self.path_str = key_sequence_to_path_ext(self.path)
        if self.path_str.endswith('.'): #root
            if isinstance(key,int):
                path = '[%d]'%(key)
            else:
                path = self.path_str + key
        else:
            if isinstance(key,int):
                path = self.path_str + '[%d]'%(key,)
            else:
                path = self.path_str + '.' + key
        self.writer.send_to_redis(path, value)

    def __delitem__(self,key):
        assert check_valid_key_name_ext(key),"{} is not a valid key under path {}".format(key,key_sequence_to_path(self.path) if self.path_str is None else self.path_str)
        if self.path_str is None:
            self.path_str = key_sequence_to_path_ext(self.path)
        if self.path_str.endswith('.'): #root
            if isinstance(key,int):
                path = '[%d]'%(key)
            else:
                path = self.path_str + key
        else:
            if isinstance(key,int):
                path = self.path_str + '[%d]'%(key,)
            else:
                path = self.path_str + '.' + key
        self.writer.delete_from_redis(path)

    def read(self):
        """Actually read the value referred to by this accessor."""
        if self.path_str is None:
            self.path_str = key_sequence_to_path_ext(self.path)
        server_value = self.reader.read_from_redis(self.path_str)
        #if it's special, then its value is under _ROOT_VALUE_READ_NAME
        try:
            return server_value[_ROOT_VALUE_READ_NAME]
        except:
            return server_value

    def write(self, value):
        """Writes value to the path referred to by this accessor."""
        if self.path_str is None:
            self.path_str = key_sequence_to_path_ext(self.path)
        self.writer.send_to_redis(self.path_str, value)

    def _do_rejson_call(self,fn,*args):
        if self.path_str is None:
            self.path_str = key_sequence_to_path_ext(self.path)
        with self.writer.interface.INTERFACE_LOCK:
            return getattr(self.writer.interface.client,fn)(self.writer.top_key_name,self.path_str,*args)

    def type(self):
        """Returns the type of the object"""
        t = self._do_rejson_call(self,'jsontype')
        return _TYPEMAP[t]

    def __len__(self):
        """Returns the length of an array / number of keys in dict"""
        try:
            return self._do_rejson_call('jsonobjlen')
        except:
            return self._do_rejson_call('jsonarrlen')

    def __iadd__(self,rhs):
        """Adds a value to an integer / float value, or concatenates a list
        to an array.

        Type checking is not performed, so the user should know what they're doing.
        """
        if not hasattr(rhs,'__iter__'):
            #treat as value
            if not isinstance(rhs,(int,float)):
                raise ValueError("+= can only accept int or float arguments")
            self._do_rejson_call('jsonnumincrby',rhs)
        else:
            self._do_rejson_call('jsonarrappend',*rhs)
        return self

    def __isub__(self,rhs):
        """Subtracts a value from an integer / float value

        Type checking is not performed, so the user should know what they're doing.
        """
        self += -rhs
        return self

    def __imul__(self,rhs):
        """Multiplies a value by an integer / float value.

        Type checking is not performed, so the user should know what they're doing.
        """
        #treat as value
        if not isinstance(rhs,(int,float)):
            raise ValueError("*= can only accept int or float arguments")
        self._do_rejson_call('jsonnummultby',rhs)
        return self

    def __idiv__(self,rhs):
        """Divides a value by an integer / float value

        Type checking is not performed, so the user should know what they're doing.
        """
        self *= 1.0/rhs
        return self

    def append(self,rhs):
        """Appends a value to an array

        Type checking is not performed, so the user should know what they're doing.
        """
        self._do_rejson_call('jsonarrappend',rhs)    




class WriteOnlyKeyAccessor(KeyAccessor):
    def __init__(self,*args,**kwargs):
        KeyAccessor.__init__(self,*args,**kwargs)
        del self.read
        del self.__delitem__
        del self.__iadd__
        del self.__isub__
        del self.__imul__
        del self.__idiv__
        del self.append


class ActiveSubscriberKeyAccessor(KeyAccessor):
    def __init__(self,*args,**kwargs):
        KeyAccessor.__init__(self,*args,**kwargs)
        del self.write
        del self.__setitem__
        del self.__delitem__
        del self.__iadd__
        del self.__isub__
        del self.__imul__
        del self.__idiv__
        del self.append

    def read(self):
        return_val = self.reader.local_copy
        dissect_path = self.path[1:]  #skip initial .
        if len(dissect_path) == 0:
            pass
        else:
            for key in dissect_path:
                return_val = return_val[key]
        if type(return_val) == dict:
            return copy_dictionary_without_paths(return_val, [])
        return return_val


class ChannelListener(Thread):
    def __init__(self, interface, channel_name, callback_function, kwargs):
        Thread.__init__(self)
        self.client = rejson.Client(host=interface.hostname)
        self.pubsub = self.client.pubsub()
        self.pubsub.psubscribe([channel_name])
        self.callback_function = callback_function
        self.kwargs = kwargs
        self.first_item_seen = False

    def run(self):
        for item in self.pubsub.listen():
            # First Item is a generic message that we need to get rid of
            if not self.first_item_seen:
                self.first_item_seen = True
                continue
            item = json_recode_str(item)
            channel = item['channel']
            message = item['data']

            self.callback_function(channel=channel, message=message, **self.kwargs)

"""
Naming:
Rapid Extendable Middleware
Redis Medium Extendable Middleware
Redis Robotic Communication

Redis Extendable Middleware

Extendable Efficient Redis Middleware
Redis Extendable Efficient Middleware
reem
"""



from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='reem',
    version='0.1.0',
    packages=['reem'],
    url='https://www.github.com/krishauser/reem',
    license='Apache 2.0',
    author='Trishul Nagenalli and Kris Hauser',
    author_email='hauser.kris@gmail.com',
    description='Redis Extendable Efficient Middleware',
    install_requires=[
        'rejson',
        'redis',
        'six',
        'numpy',
      ],
    long_description=long_description,
    long_description_content_type='text/markdown',
)
